import logging

import requests
from cashaddress import convert as cashaddress

from bitcash.network import currency_to_satoshi
from bitcash.network.meta import Unspent

from collections import OrderedDict

from datetime import datetime,timezone,date

DEFAULT_TIMEOUT = 30


def set_service_timeout(seconds):
    global DEFAULT_TIMEOUT
    DEFAULT_TIMEOUT = seconds

class InsightAPI:
    MAIN_ENDPOINT = ''
    # MAIN_ADDRESS_API = ''
    # MAIN_BALANCE_API = ''
    # MAIN_UNSPENT_API = ''
    # MAIN_TX_PUSH_API = ''
    # MAIN_TX_AMOUNT_API = ''
    # TX_PUSH_PARAM = ''

    MAIN_ADDRESS_API = 'addr/'
    MAIN_BALANCE_API = 'addr/{}/balance'
    MAIN_UNSPENT_API = 'addr/{}/utxo'
    MAIN_TX_PUSH_API = 'tx/send'
    MAIN_TX_API = 'tx/'
    MAIN_RAWTX_API='rawtx/'
    # MAIN_TXS_BY_ADDRESS_API= 'txs/?address='  # Only can return first 10 txs
    MAIN_TXS_BY_ADDRESSES_API='addrs/{}/txs?from={}&to={}'
    TX_PUSH_PARAM = 'rawtx'

    MAIN_TXS_BY_BLOCK='txs/?block={}&pageNum={}'
    MAIN_BLOCK_SUMMARIES_BY_DATE='blocks?blockDate={}'  # date: 2016-4-20

    MAIN_BLOCKHASH_BY_HEIGHT= 'block-index/'


    NEW_ADDRESS_SUPPORTED=True

    @classmethod
    def get_balance(cls, address):
        if not cls.NEW_ADDRESS_SUPPORTED:
            address = cashaddress.to_legacy_address(address)
        r = requests.get((cls.MAIN_ENDPOINT+cls.MAIN_BALANCE_API).format(address), timeout=DEFAULT_TIMEOUT)
        if r.status_code != 200:  # pragma: no cover
            raise ConnectionError
        return r.json()

    @classmethod
    def get_transactions(cls, address):
        """

        :param address:
        :return:list of transaction hash
        """
        if not cls.NEW_ADDRESS_SUPPORTED:
            address = cashaddress.to_legacy_address(address)
        r = requests.get(cls.MAIN_ENDPOINT+cls.MAIN_ADDRESS_API + address, timeout=DEFAULT_TIMEOUT)
        if r.status_code != 200:  # pragma: no cover
            raise ConnectionError
        return r.json()['transactions']

    @classmethod
    def get_tx_amount(cls, txid, txindex):
        r = requests.get(cls.MAIN_ENDPOINT+cls.MAIN_TX_API+txid, timeout=DEFAULT_TIMEOUT)
        if r.status_code != 200:  # pragma: no cover
            raise ConnectionError
        response = r.json()
        return response['vout'][txindex]['value']*100000000

    @classmethod
    def get_tx(cls, txid):
        r = requests.get(cls.MAIN_ENDPOINT+cls.MAIN_TX_API+txid, timeout=DEFAULT_TIMEOUT)
        if r.status_code != 200:  # pragma: no cover
            raise ConnectionError
        response = r.json()
        return response

    @classmethod
    def get_rawtx(cls, txid):
        r = requests.get(cls.MAIN_ENDPOINT+cls.MAIN_RAWTX_API+txid, timeout=DEFAULT_TIMEOUT)
        if r.status_code != 200:  # pragma: no cover
            raise ConnectionError
        response = r.json()
        return response['rawtx']

    @classmethod
    def get_unspent(cls, address):
        if not cls.NEW_ADDRESS_SUPPORTED:
            address = cashaddress.to_legacy_address(address)
        r = requests.get((cls.MAIN_ENDPOINT+cls.MAIN_UNSPENT_API).format(address), timeout=DEFAULT_TIMEOUT)
        if r.status_code != 200:  # pragma: no cover
            raise ConnectionError
        return [
            Unspent(currency_to_satoshi(tx['amount'], 'bch'),
                    tx['confirmations'],
                    tx['scriptPubKey'],
                    tx['txid'],
                    tx['vout'])
            for tx in r.json()
        ]

    @classmethod
    def broadcast_tx(cls, tx_hex):  # pragma: no cover
        r = requests.post(cls.MAIN_ENDPOINT+cls.MAIN_TX_PUSH_API, data={cls.TX_PUSH_PARAM: tx_hex}, timeout=DEFAULT_TIMEOUT)
        if r.status_code!=200:
            raise Exception(r.content)
        return True if r.status_code == 200 else False

    @classmethod
    def get_transactions_by_addresses(cls,addresses,start_index=0,stop_index=50):
        if stop_index-start_index>50:
            raise ValueError('Range between start_index ({}) and stop_index '
                             '({}) less than or equal to 50!'.format(start_index,stop_index))
        if isinstance(addresses,str):
            addresses=[addresses]
        if not cls.NEW_ADDRESS_SUPPORTED:
            addresses = [cashaddress.to_legacy_address(address) for address in addresses]
        addresses_str=','.join(addresses)

        r=requests.get((cls.MAIN_ENDPOINT + cls.MAIN_TXS_BY_ADDRESSES_API).
                       format(addresses_str,start_index,stop_index), timeout=DEFAULT_TIMEOUT)
        if r.status_code!=200:
            raise ConnectionError

        data=r.json()
        total_txs=data['totalItems']
        txs=data['items']   #type:list
        return total_txs,txs

    @classmethod
    def get_all_transactions_by_address(cls,address):
        """
        Get all txs related to an address
        :param address:
        :return:
        """
        total_txs,txs=cls.get_transactions_by_addresses(address)
        if total_txs<=50:
            return txs
        else:
            # if total_txs>50*10:
            #     raise TooManyTxs('Too many txs!')
            #     # todo: deal with too many txs
            start_indexes=list(range(50,total_txs,50))
            stop_indexes=list(range(100,total_txs,50))+[total_txs]
            for start_index,stop_index in zip(start_indexes,stop_indexes):
                total_txs, txs_follow_up = cls.get_transactions_by_addresses\
                    (address,start_index=start_index,stop_index=stop_index)
                txs=txs+txs_follow_up

            txs_id=[tx['txid'] for tx in txs]
            normal_dict=dict(zip(txs_id,txs))
            txs_id_no_duplicate=list(OrderedDict.fromkeys(txs_id).keys())  # remove duplicate txs
            txs_no_duplicate=[normal_dict[txid] for txid in txs_id_no_duplicate]
            return txs_no_duplicate

    @classmethod
    def get_transactions_by_address_from(cls,address,t):
        """
        Get all txs related to an address after t ,some tx after t may included
        :param address:
        :param timestap
        :return:
        """
        total_txs,txs=cls.get_transactions_by_addresses(address)
        if int(txs[-1]['time'])<t:
            return txs

        if total_txs<=50:
            return txs
        else:
            # if total_txs>50*10:
            #     raise TooManyTxs('Too many txs!')
            #     # todo: deal with too many txs
            start_indexes=list(range(50,total_txs,50))
            stop_indexes=list(range(100,total_txs,50))+[total_txs]
            for start_index,stop_index in zip(start_indexes,stop_indexes):
                total_txs, txs_follow_up = cls.get_transactions_by_addresses\
                    (address,start_index=start_index,stop_index=stop_index)
                txs=txs+txs_follow_up
                if int(txs[-1]['time']) < t:
                    break

            txs_id=[tx['txid'] for tx in txs]
            normal_dict=dict(zip(txs_id,txs))
            txs_id_no_duplicate=list(OrderedDict.fromkeys(txs_id).keys())  # remove duplicate txs
            txs_no_duplicate=[normal_dict[txid] for txid in txs_id_no_duplicate]
            return txs_no_duplicate

    @classmethod
    def get_transactions_by_block(cls,block_hash,page_num=0):
        r=requests.get((cls.MAIN_ENDPOINT+cls.MAIN_TXS_BY_BLOCK).format(block_hash,page_num),timeout=DEFAULT_TIMEOUT)
        if r.status_code!=200:
            raise  ConnectionError
        data=r.json()
        pages_total=int(data['pagesTotal'])
        txs=data['txs']  # type:list
        return pages_total,txs

    @classmethod
    def get_all_transactions_by_block(cls,block_hash):
        pages_total,txs=cls.get_transactions_by_block(block_hash,0)
        for i in range(1,pages_total):
            pt,txs_new=cls.get_transactions_by_block(block_hash,i)
            txs+=txs_new
        return txs

    @classmethod
    def get_block_summaries_by_date(cls,datestr=None):
        """

        :param datestr:2016-06-12
        :return:
        """
        if datestr is None:
            datestr=datetime.now().astimezone(timezone.utc).date().isoformat()
        
        r=requests.get((cls.MAIN_ENDPOINT+cls.MAIN_BLOCK_SUMMARIES_BY_DATE).format(datestr),timeout=DEFAULT_TIMEOUT)
        if r.status_code!=200:
            raise  ConnectionError
        data=r.json()
        block_summaries=data['blocks']
        return list(reversed(block_summaries))

    @ classmethod
    def get_block_summaries_by_from_to(cls,t_start,t_stop):
        """

        :param t_start: Unix timestamp
        :param t_stop:  Unix timestamp
        :return: block summaries that block time>=t_start and block.time<=t_stop
        """
        date_start=datetime.fromtimestamp(t_start,timezone.utc).date()
        date_stop=datetime.fromtimestamp(t_stop,timezone.utc).date()

        dates_ordinal_list=list(range(date_start.toordinal(),date_stop.toordinal()))+[(date_stop.toordinal())]
        dates=[date.fromordinal(day) for day in dates_ordinal_list]

        block_summaries=[]
        for day in dates:
            block_summaries+=cls.get_block_summaries_by_date(day.isoformat())

        block_summaries=[block_summary for block_summary in block_summaries
                         if int(block_summary['time'])>=t_start and int(block_summary['time']<=t_stop)]
        return block_summaries

    @classmethod
    def get_transactions_from_to(cls,t_start,t_stop):
        block_summaries=cls.get_block_summaries_by_from_to(t_start,t_stop)
        txs=[]
        for block in block_summaries:
            txs+=cls.get_all_transactions_by_block(block['hash'])
        return txs

    @classmethod
    def get_blockhash_by_heigth(cls,height):
        r=requests.get(cls.MAIN_ENDPOINT+ cls.MAIN_BLOCKHASH_BY_HEIGHT+str(height))
        if r.status_code!=200:
            raise  ConnectionError
        data=r.json()
        return data['blockHash']


class BCCBlockAPI(InsightAPI):
    """
    bccblock.info
    No testnet, sadly. Also uses legacy addresses only.
    """
    MAIN_ENDPOINT = 'https://bccblock.info/api/'
    # MAIN_ADDRESS_API = MAIN_ENDPOINT + 'addr/'
    # MAIN_BALANCE_API = MAIN_ADDRESS_API + '{}/balance'
    # MAIN_UNSPENT_API = MAIN_ADDRESS_API + '{}/utxo'
    # MAIN_TX_PUSH_API = MAIN_ENDPOINT + 'tx/send'
    # MAIN_TX_AMOUNT_API = MAIN_ENDPOINT + 'tx'
    # TX_PUSH_PARAM = 'rawtx'

    NEW_ADDRESS_SUPPORTED = False

    # @classmethod
    # def get_balance(cls, address):
    #     # As of 2018-02-02, bccblock.info only supports legacy addresses.
    #     address = cashaddress.to_legacy_address(address)
    #     r = requests.get(cls.MAIN_BALANCE_API.format(address), timeout=DEFAULT_TIMEOUT)
    #     if r.status_code != 200:  # pragma: no cover
    #         raise ConnectionError
    #     return r.json()

    # @classmethod
    # def get_transactions(cls, address):
    #     # As of 2018-02-02, bccblock.info only supports legacy addresses.
    #     address = cashaddress.to_legacy_address(address)
    #     r = requests.get(cls.MAIN_ADDRESS_API + address, timeout=DEFAULT_TIMEOUT)
    #     if r.status_code != 200:  # pragma: no cover
    #         raise ConnectionError
    #     return r.json()['transactions']

    # @classmethod
    # def get_unspent(cls, address):
    #     # As of 2018-02-02, bccblock.info only supports legacy addresses.
    #     address = cashaddress.to_legacy_address(address)
    #     r = requests.get(cls.MAIN_UNSPENT_API.format(address), timeout=DEFAULT_TIMEOUT)
    #     if r.status_code != 200:  # pragma: no cover
    #         raise ConnectionError
    #     return [
    #         Unspent(currency_to_satoshi(tx['amount'], 'bch'),
    #                 tx['confirmations'],
    #                 tx['scriptPubKey'],
    #                 tx['txid'],
    #                 tx['vout'])
    #         for tx in r.json()
    #     ]

class BlockdozerAPI(InsightAPI):
    MAIN_ENDPOINT = 'https://blockdozer.com/insight-api/'
    # MAIN_ADDRESS_API = MAIN_ENDPOINT + 'addr/'
    # MAIN_BALANCE_API = MAIN_ADDRESS_API + '{}/balance'
    # MAIN_UNSPENT_API = MAIN_ADDRESS_API + '{}/utxo'
    # MAIN_TX_PUSH_API = MAIN_ENDPOINT + 'tx/send'
    # MAIN_TX_AMOUNT_API = MAIN_ENDPOINT + 'tx'
    TEST_ENDPOINT = 'https://tbch.blockdozer.com/insight-api/'
    TEST_ADDRESS_API = TEST_ENDPOINT + 'addr/'
    TEST_BALANCE_API = TEST_ADDRESS_API + '{}/balance'
    TEST_UNSPENT_API = TEST_ADDRESS_API + '{}/utxo'
    TEST_TX_PUSH_API = TEST_ENDPOINT + 'tx/send'
    # TX_PUSH_PARAM = 'rawtx'

    NEW_ADDRESS_SUPPORTED = True

    @classmethod
    def get_balance_testnet(cls, address):
        r = requests.get(cls.TEST_BALANCE_API.format(address), timeout=DEFAULT_TIMEOUT)
        if r.status_code != 200:  # pragma: no cover
            raise ConnectionError
        return r.json()

    @classmethod
    def get_transactions_testnet(cls, address):
        r = requests.get(cls.TEST_ADDRESS_API + address, timeout=DEFAULT_TIMEOUT)
        if r.status_code != 200:  # pragma: no cover
            raise ConnectionError
        return r.json()['transactions']

    @classmethod
    def get_unspent_testnet(cls, address):
        r = requests.get(cls.TEST_UNSPENT_API.format(address), timeout=DEFAULT_TIMEOUT)
        if r.status_code != 200:  # pragma: no cover
            raise ConnectionError
        return [
            Unspent(currency_to_satoshi(tx['amount'], 'bch'),
                    tx['confirmations'],
                    tx['scriptPubKey'],
                    tx['txid'],
                    tx['vout'])
            for tx in r.json()
        ]

    @classmethod
    def broadcast_tx_testnet(cls, tx_hex):  # pragma: no cover
        r = requests.post(cls.TEST_TX_PUSH_API, data={cls.TX_PUSH_PARAM: tx_hex}, timeout=DEFAULT_TIMEOUT)
        if r.status_code == 200:
            return True
        else:
            logging.error(r.text)
            return False

class NetworkAPI:
    IGNORED_ERRORS = (ConnectionError,
                      requests.exceptions.ConnectionError,
                      requests.exceptions.Timeout,
                      requests.exceptions.ReadTimeout)

    GET_BALANCE_MAIN = [BCCBlockAPI.get_balance,
                        BlockdozerAPI.get_balance]
    GET_TRANSACTIONS_MAIN = [BCCBlockAPI.get_transactions,
                             BlockdozerAPI.get_transactions]
    GET_UNSPENT_MAIN = [BCCBlockAPI.get_unspent,
                        BlockdozerAPI.get_unspent]
    BROADCAST_TX_MAIN = [BCCBlockAPI.broadcast_tx,
                         BlockdozerAPI.broadcast_tx]
    GET_TX_MAIN= [BCCBlockAPI.get_tx,
                  BlockdozerAPI.get_tx]
    GET_RAWTX_MAIN= [BCCBlockAPI.get_rawtx,
                  BlockdozerAPI.get_rawtx]
    GET_TX_AMOUNT = [BCCBlockAPI.get_tx_amount,
                     BlockdozerAPI.get_tx_amount]
    GET_ALL_TXS_BY_ADDRESS=[BCCBlockAPI.get_all_transactions_by_address,
                        BlockdozerAPI.get_all_transactions_by_address]

    GET_TXS_BY_ADDRESSES=[BCCBlockAPI.get_transactions_by_addresses,
                          BlockdozerAPI.get_transactions_by_addresses]

    GET_TXS_BY_ADDRESS_FROM=[BCCBlockAPI.get_transactions_by_address_from,
                        BlockdozerAPI.get_transactions_by_address_from]

    BLOCKHASH_BY_HEIGHT=[BCCBlockAPI.get_blockhash_by_heigth,
                         BlockdozerAPI.get_blockhash_by_heigth]

    GET_BALANCE_TEST = [BlockdozerAPI.get_balance_testnet]
    GET_TRANSACTIONS_TEST = [BlockdozerAPI.get_transactions_testnet]
    GET_UNSPENT_TEST = [BlockdozerAPI.get_unspent_testnet]
    BROADCAST_TX_TEST = [BlockdozerAPI.broadcast_tx_testnet]

    @classmethod
    def get_balance(cls, address):
        """Gets the balance of an address in satoshi.

        :param address: The address in question.
        :type address: ``str``
        :raises ConnectionError: If all API services fail.
        :rtype: ``int``
        """

        for api_call in cls.GET_BALANCE_MAIN:
            try:
                return api_call(address)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def get_balance_testnet(cls, address):
        """Gets the balance of an address on the test network in satoshi.

        :param address: The address in question.
        :type address: ``str``
        :raises ConnectionError: If all API services fail.
        :rtype: ``int``
        """

        for api_call in cls.GET_BALANCE_TEST:
            try:
                return api_call(address)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def get_transactions(cls, address):
        """Gets the ID of all transactions related to an address.

        :param address: The address in question.
        :type address: ``str``
        :raises ConnectionError: If all API services fail.
        :rtype: ``list`` of ``str``
        """

        for api_call in cls.GET_TRANSACTIONS_MAIN:
            try:
                return api_call(address)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def get_transactions_testnet(cls, address):
        """Gets the ID of all transactions related to an address on the test
        network.

        :param address: The address in question.
        :type address: ``str``
        :raises ConnectionError: If all API services fail.
        :rtype: ``list`` of ``str``
        """

        for api_call in cls.GET_TRANSACTIONS_TEST:
            try:
                return api_call(address)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def get_tx_amount(cls, txid, txindex):
        """Gets the ID of all transactions related to an address.

        :param txid: The transaction id in question.
        :type txid: ``str``
        :param txindex: The transaction index in question.
        :type txindex: ``str``
        :raises ConnectionError: If all API services fail.
        :rtype: ``list`` of ``str``
        """

        for api_call in cls.GET_TX_AMOUNT:
            try:
                return api_call(txid, txindex)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def get_tx(cls, txid):
        """Gets tx dict by txid.

        :param txid: The transaction id in question.
        :type txid: ``str``
        :raises ConnectionError: If all API services fail.
        :rtype: ``dict``
        """

        for api_call in cls.GET_TX_MAIN:
            try:
                return api_call(txid)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def get_rawtx(cls, txid):
        """Gets rawtx by txid.

        :param txid: The transaction id in question.
        :type txid: ``str``
        :raises ConnectionError: If all API services fail.
        :rtype: ``str``
        """

        for api_call in cls.GET_RAWTX_MAIN:
            try:
                return api_call(txid)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def get_unspent(cls, address):
        """Gets all unspent transaction outputs belonging to an address.

        :param address: The address in question.
        :type address: ``str``
        :raises ConnectionError: If all API services fail.
        :rtype: ``list`` of :class:`~bitcash.network.meta.Unspent`
        """

        for api_call in cls.GET_UNSPENT_MAIN:
            try:
                return api_call(address)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def get_transactions_by_addresses(cls,addresses):
        """Gets latest 50 transactions in dict related to an address.

        :param addresses: bch addresse(s)
        :type addresses: ``str`` or ''list'' of ''str''
        :raises ConnectionError: If all API services fail.
        :rtype: ``list'' of ``dict``
        :return: All transactions in dict related to the address
        """

        for api_call in cls.GET_TXS_BY_ADDRESSES:
            try:
                return api_call(addresses)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def get_all_transactions_by_address(cls,address):
        """Gets all transactions in dict related to an address.

        :param address: bch address
        :type address: ``str``
        :raises ConnectionError: If all API services fail.
        :rtype: ``list'' of ``dict``
        :return: All transactions in dict related to the address
        """

        for api_call in cls.GET_ALL_TXS_BY_ADDRESS:
            try:
                return api_call(address)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def get_transactions_by_address_from(cls,address,t):
        """Gets all transactions in dict related to an address from time t

        :param address: bch address
        :type address: ``str``
        :raises ConnectionError: If all API services fail.
        :rtype: ``list'' of ``dict``
        :return: All transactions in dict related to the address
        """

        for api_call in cls.GET_TXS_BY_ADDRESS_FROM:
            try:
                return api_call(address,t)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def get_blockhash_by_height(cls,height):
        for api_call in cls.BLOCKHASH_BY_HEIGHT:
            try:
                return api_call(height)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')


    @classmethod
    def get_unspent_testnet(cls, address):
        """Gets all unspent transaction outputs belonging to an address on the
        test network.

        :param address: The address in question.
        :type address: ``str``
        :raises ConnectionError: If all API services fail.
        :rtype: ``list`` of :class:`~bitcash.network.meta.Unspent`
        """

        for api_call in cls.GET_UNSPENT_TEST:
            try:
                return api_call(address)
            except cls.IGNORED_ERRORS:
                pass

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def broadcast_tx(cls, tx_hex):  # pragma: no cover
        """Broadcasts a transaction to the blockchain.

        :param tx_hex: A signed transaction in hex form.
        :type tx_hex: ``str``
        :raises ConnectionError: If all API services fail.
        """
        success = None

        for api_call in cls.BROADCAST_TX_MAIN:
            try:
                success = api_call(tx_hex)
                if not success:
                    continue
                return
            except cls.IGNORED_ERRORS:
                pass

        if success is False:
            raise ConnectionError('Transaction broadcast failed, or '
                                  'Unspents were already used.')

        raise ConnectionError('All APIs are unreachable.')

    @classmethod
    def broadcast_tx_testnet(cls, tx_hex):  # pragma: no cover
        """Broadcasts a transaction to the test network's blockchain.

        :param tx_hex: A signed transaction in hex form.
        :type tx_hex: ``str``
        :raises ConnectionError: If all API services fail.
        """
        success = None

        for api_call in cls.BROADCAST_TX_TEST:
            try:
                success = api_call(tx_hex)
                if not success:
                    continue
                return
            except cls.IGNORED_ERRORS:
                pass

        if success is False:
            raise ConnectionError('Transaction broadcast failed, or '
                                  'Unspents were already used.')

        raise ConnectionError('All APIs are unreachable.')

class TooManyTxs(Exception):
    pass


from bitcash.wallet import PrivateKey
from cashaddress.convert import is_valid, to_cash_address

from bchmemo.bitcash_modified.services import NetworkAPI
from bchmemo.memo import Memo
from bchmemo.memo import PRIFIX_BY_ACTION_NAME
from bchmemo.memo import USER_NAME_DICT
from bchmemo.memo import get_name_from_address

PROMPT=True  # Prompt or not

class MemoUser:
    """
    This class represents  a memo user and provides functions about read and send memos

    :param address: the address of user
    :type address: ``str``

    """

    def __init__(self,address):
        if not isinstance(address,str):
            raise TypeError('address should be str!')
        if not is_valid(address):
            raise ValueError('{} is not a valid address'.format(address))

        self._address = to_cash_address(address)
        self.memos_send=[]  # list of memos sent by user
        self.memos_post=[]  # list of post memo sent by user
        self.memos_like=[]  # list of like / tip memo sent by user
        self.memos_reply=[]

        self.memos_receive=[]   # list of memos that are not sent by user but transfer BCH to user

        self._private_key=None

        self.name=None
        self.following=set()

    @classmethod
    def from_private_key(cls,private_key):
        """
        Return a memo user from private key.

        :param private_key:
        :type private_key: ``str`` or ``PrivateKey``
        :rtype: MemoUser
        """
        if not isinstance(private_key,PrivateKey):
            pk=PrivateKey(private_key)
        else:
            pk=private_key
        memouser=MemoUser(pk.address)
        memouser.private_key=pk.to_wif()
        return memouser

    def get_memos(self):
        """
        Get 1)memos sent by user, 2)memos that transfer BCH to user (tip memos to user)
        """
        total_txs,txs=NetworkAPI.get_transactions_by_addresses(self._address)
        memos=[Memo.form_transaction_dict(tx) for tx in txs if Memo.is_memo(tx)]
        self.memos_send=[memo for memo in memos if memo.sender==self._address]
        self.memos_receive=[memo for memo in memos if memo.sender!=self._address]

        self.memos_post=[memo for memo in self.memos_send if memo.prefix == PRIFIX_BY_ACTION_NAME['Post memo']]
        self.memos_like=[memo for memo in self.memos_send if memo.prefix == PRIFIX_BY_ACTION_NAME['Like / tip memo']]
        self.memos_reply=[memo for memo in self.memos_send if memo.prefix == PRIFIX_BY_ACTION_NAME['Reply to memo']]
        for memo in reversed(self.memos_send):
            if memo.prefix==PRIFIX_BY_ACTION_NAME['Set name']:
                self.name=memo.name
                USER_NAME_DICT[memo.sender]=memo.name
            if memo.prefix==PRIFIX_BY_ACTION_NAME['Follow user']:
                self.following.add(memo.address)
            if memo.prefix==PRIFIX_BY_ACTION_NAME['Unfollow user']:
                self.following.remove(memo.address)

    def get_memos_from(self,t):
        """

        """
        txs=NetworkAPI.get_transactions_by_address_from(self._address,t)
        memos=[Memo.form_transaction_dict(tx) for tx in txs if Memo.is_memo(tx)]
        self.memos_send=[memo for memo in memos if memo.sender==self._address]
        self.memos_receive=[memo for memo in memos if memo.sender!=self._address]

        self.memos_post=[memo for memo in self.memos_send if memo.prefix == PRIFIX_BY_ACTION_NAME['Post memo']]
        self.memos_like=[memo for memo in self.memos_send if memo.prefix == PRIFIX_BY_ACTION_NAME['Like / tip memo']]
        for memo in reversed(self.memos_send):
            if memo.prefix==PRIFIX_BY_ACTION_NAME['Set name']:
                self.name=memo.name
                USER_NAME_DICT[memo.sender]=memo.name
            if memo.prefix==PRIFIX_BY_ACTION_NAME['Follow user']:
                self.following.add(memo.address)
            if memo.prefix==PRIFIX_BY_ACTION_NAME['Unfollow user']:
                self.following.remove(memo.address)

    def list_posts(self):
        for memo in self.memos_post:
            print(memo.content_post())

    def list_likes(self):
        for memo in self.memos_like:
            print(memo.content_like())

    def list_following(self):
        print('following:')
        for addr in list(self.following):
            print(get_name_from_address(addr),addr)

    def list_replies(self):
        for memo in self.memos_reply:
            print(memo.content_reply())

    def list_memos(self):
        self.list_posts()
        self.list_replies()
        self.list_likes()
        print()
        self.list_following()

    @property
    def private_key(self):
        return self._private_key

    @private_key.setter
    def private_key(self,private_key):
        pk=PrivateKey(private_key)
        if pk.address != self._address:
            raise ValueError('Wrong private key!')
        self._private_key=private_key

    def __send_new_memo(self,memo):
        """

        :param memo:
        :type memo: Memo
        :return:
        """
        memo.create_signed_transaction(self._private_key)
        txid=memo.send_transaction()
        if PROMPT:
            print('Successfully sent! txid={}'.format(txid))
            print('Check it on: https://explorer.bitcoin.com/bch/tx/{}'.format(txid))
        return txid

    def set_name(self, name):
        """
        :param name: no more than 76 bytes
        :type name: ``str``
        :return: transaction id
        """
        memo=Memo.set_name(name,self._address)
        return self.__send_new_memo(memo)

    def post_memo(self, message):
        """
        :param message: no more than 76 bytes
        :type message: ``str``
        :return: transaction id
        """
        memo=Memo.post_memo(message,self._address)
        return self.__send_new_memo(memo)

    def like_memo(self, liked_memo, tip_amount=None, sender_of_liked_memo=None):
        """
        :param liked_memo: the memo or to tip or like, or, the transaction hash of that memo
        :type liked_memo: ``Memo`` or ``str``
        :param tip_amount: 0 for like, >546 for tip, in satoshi
        :type tip_amount: ``int``
        :param sender_of_liked_memo: sender's address of liked memo.If liked_memo is transaction
                    hash and this is None, this need be get from NetworkAPI
        :type: sender_of_liked_memo: ``str``
        :return:
        """
        if isinstance(liked_memo,Memo):
            memo = Memo.like_memo(liked_memo, self._address, tip_amount=tip_amount)
        elif isinstance(liked_memo,str):
            if sender_of_liked_memo is None:
                liked_memo=Memo.from_txhash(liked_memo)
                memo = Memo.like_memo(liked_memo, self._address, tip_amount=tip_amount)
            else:
                memo = Memo.like_memo(liked_memo, self._address, tip_amount=tip_amount, sender_of_liked_memo=sender_of_liked_memo)
        else:
            raise TypeError('liked_memo should be Memo object or string!')

        return self.__send_new_memo(memo)

    def reply_to_memo(self,target_memo,message):
        """

        :param target_memo:
        :type target_memo: ``Memo`` or ``str``
        :param message: no more than 41 bytes
        :type message: ``str``
        :return
        """
        memo=Memo.reply_to_memo(target_memo,message,self._address)
        return self.__send_new_memo(memo)

    def follow(self, address):
        """
        :param address:
        :type address: ``str``
        :return: transaction id
        """

        memo=Memo.follow(address,self._address)
        return self.__send_new_memo(memo)

    def unfollow(self, address):
        """
        :param address: no more than 76 bytes
        :type address: ``str``
        :return: transaction id
        """
        memo=Memo.unfollow(address,self._address)
        return self.__send_new_memo(memo)

class UnknownPrivateKey(Exception):
    pass



from datetime import datetime
import re
import sys

from bitcash.format import address_to_public_key_hash
from cashaddress.convert import Address
from cashaddress.convert import to_cash_address

from bchmemo.bitcash_modified.services import NetworkAPI

import bitcash.wallet
import bchmemo.bitcash_modified.transaction as transaction_modified

# modify objects related to transaction in bitcash.wallet
attr_to_be_modified='calc_txid, create_p2pkh_transaction, sanitize_tx_data, ' \
                    'OP_CHECKSIG, OP_DUP, OP_EQUALVERIFY, OP_HASH160, OP_PUSH_20'
attr_to_be_modified=re.split('\W+',attr_to_be_modified)
for attr in attr_to_be_modified:
    setattr(bitcash.wallet,attr,getattr(transaction_modified,attr))

# Parameters
PrivateKey=bitcash.wallet.PrivateKey

SUPPORTED_PREFIX=['6d01','6d02','6d04','6d06','6d07','6d03']
ACTION_NAME=['Set name','Post memo','Like / tip memo','Follow user','Unfollow user','Reply to memo']
PRIFIX_BY_ACTION_NAME=dict(zip(ACTION_NAME,SUPPORTED_PREFIX))

USER_NAME_DICT={}

MIN_TRANSFER_FEE=1 # satoshi per Byte

class Memo:
    """
    This class represents a single memo and provides function including
    analyzing memo from a transaction and creating a transaction with memo.
    """

    def __init__(self):
        self.__action_name=None
        self.sender=''  # sender address
        self._prefix=None # hex string
        self._values= '' # hex string in OP_RETURN data except prefix


        self._name=''
        self._message=''
        self._txhash_of_target_memo= ''  # txid of the memo that is be liked or replied
        self._sender_of_liked_memo=''
        self.liked_memo=None
        self._tip_amount=0
        self._address=''  # hex string of public key hash

        self.transaction_hash=None
        self.transaction_dict=None
        self.transaction_time=None
        self.blockheight=None

        self.signed_transaction=None

        self.is_parsed_correctly=True
        self.parsed_error=None

        self.transfer=[]

    @classmethod
    def is_memo(cls,transaction:dict):
        """Return whether or not a transaction (dict format) is a memo transaction.

        :param  transaction: a dict format transaction
        :type transaction: dict

        """
        for vout in transaction['vout']:
            if vout['scriptPubKey']['hex'][0:2] == '6a' \
                    and vout['scriptPubKey']['hex'][4:8] in SUPPORTED_PREFIX:
                return True
        else:
            return False

    @classmethod
    def form_transaction_dict(cls, transaction):
        if not cls.is_memo(transaction):
            raise ValueError('Not valid transaction dict!')

        memo = Memo()

        try:
            memo.transaction_dict=transaction
            memo.transaction_hash=transaction['txid']
            memo.blockheight=transaction['blockheight']
            # memo.transaction_time=transaction['time']

            memo.__get_transfer()
            for vout in transaction['vout']:
                if vout['scriptPubKey']['hex'][0:2] == '6a' \
                        and vout['scriptPubKey']['hex'][4:8] in SUPPORTED_PREFIX:
                    memo.prefix= vout['scriptPubKey']['hex'][4:8]
                    memo.values= vout['scriptPubKey']['hex'][8:]
                    break
            else:
                raise NotMemoTransaction('No OP_RETURN data in the transaction or none memo supported data.')

            addr_vin = []
            for vin in transaction['vin']:  # type:dict
                if 'addr' in vin:
                    addr_vin.append(to_cash_address(vin['addr']))
            if len(addr_vin)>=1:
                memo.sender=addr_vin[0]

            memo.transaction_time=transaction['time']

        except:
            memo.is_parsed_correctly=False
            memo.parsed_error=sys.exc_info()[1]

        return memo

    def __get_transfer(self):
        """
        generate transfer form transaction dict
        """
        transfer=[]
        for vout in self.transaction_dict['vout']:
            if 'addresses' in vout['scriptPubKey']:
                transfer.append((to_cash_address(vout['scriptPubKey']['addresses'][0]),vout['value']))
        self.transfer=transfer

    @property
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self,prefix):
        self._prefix=prefix
        try:
            index=SUPPORTED_PREFIX.index(prefix)
        except ValueError:
            raise ValueError('"{}" is not a supported memo prefix!',self._prefix)
        self.__action_name=ACTION_NAME[index]

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self,value):
        """
            :param value: hex string in OP_RETURN data except prefix and length bytes
        """
        value_bytes=bytes.fromhex(value)
        if self._prefix==PRIFIX_BY_ACTION_NAME['Set name']:
            self.name=value_bytes[1:].decode()
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Post memo']:
            self.message=value_bytes[1:].decode()
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Like / tip memo']:
            self.txhash_of_target_memo=bytes(reversed(value_bytes[1:1+32])).hex()
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Follow user']:
            self.address=value_bytes[1:].hex()
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Unfollow user']:
            self.address=value_bytes[1:].hex()
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Reply to memo']:
            self.txhash_of_target_memo= bytes(reversed(value_bytes[1:1 + 32])).hex()
            self.message=value_bytes[34:].decode()
        else:
            raise ValueError('"{}" is not a supported memo prefix!',self._prefix)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self,name):
        if not isinstance(name,str):
            raise TypeError('name should be a string!')
        if len(name)>76:
            raise ValueError('"{}" is too long.Max length of memo name is 76 bytes',name)
        else:
            self._name=name

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, message):
        if not isinstance(message,str):
            raise TypeError('message should be a string!')
        if self._prefix==PRIFIX_BY_ACTION_NAME['Post memo'] and len(message)>76:
            self._message = message
            raise ValueError('"{}" is too long.Max length of memo message is 76 bytes',message)
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Reply to memo'] and len(message)>41:  #todo: check message lenth
            self._message = message
            raise ValueError('"{}" is too long.Max length of memo message is 76 bytes',message)
        else:
            self._message=message

    @property
    def txhash_of_target_memo(self):
        return self._txhash_of_target_memo

    @txhash_of_target_memo.setter
    def txhash_of_target_memo(self, txhash):
        if not isinstance(txhash,str):
            raise TypeError('txhash should be a string!')
        if len(txhash)!=64:
            raise ValueError('txhash("{}") should be 64 hex chars long'.format(txhash))
        else:
            self._txhash_of_target_memo=txhash
            # self.__get_like_memo()

    @property
    def sender_of_liked_memo(self):
        return self._sender_of_liked_memo

    @sender_of_liked_memo.setter
    def sender_of_liked_memo(self,sender):
        self._sender_of_liked_memo=sender
        for addr, amount in self.transfer:
            if addr == sender:
                self._tip_amount += int(amount.replace('.',''))

    def __get_like_memo(self):
        """
        Get memo liked by this memo and calc tip amount
        """
        try:
            tx_like=NetworkAPI.get_tx(self._txhash_of_target_memo)
        except:
            return
        if Memo.is_memo(tx_like):
            self.liked_memo=Memo.form_transaction_dict(tx_like)  # FIXME: Long runtime for multi-nesting like type memo
            for addr,amount in self.transfer:
                if addr==self.liked_memo.sender:
                    self._tip_amount+=int(float(amount) * 100000000)

    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, address):
        if not isinstance(address,str):
            raise TypeError('address should be a string!')
        if len(address)!=40:
            self._address=''
            raise ValueError('"{}" is not a public key hash'.format(address))
        else:
            self._address=Address(payload=list(bytes.fromhex(address[0:40])),version= 'P2PKH').cash_address()

    def content(self):
        if self.sender and self._prefix and self._values:
            return 'EMPTY MEMO'

        line_sender=get_name_from_address(self.sender)+'@...'+self.sender[-6:]\
                    +' posted at '+datetime.fromtimestamp(int(self.transaction_time)).strftime('%Y-%m-%d %H:%M:%S')

        line_message=''
        if self._prefix==PRIFIX_BY_ACTION_NAME['Set name']:
            line_message= self.__action_name + ': ' + self.name
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Post memo']:
            line_message = self.__action_name + ': ' + self.message
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Like / tip memo']:
            line_message = self.__action_name + ': ' + self.txhash_of_target_memo
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Follow user']:
            line_message = self.__action_name + ': ' + Address('P2PKH', list(bytes.fromhex(self.address))).cash_address()
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Unfollow user']:
            line_message = self.__action_name + ': ' + Address('P2PKH', list(bytes.fromhex(self.address))).cash_address()

        return line_sender+'\n'+line_message+'\n'

    def content_post(self):
        if self.prefix!=PRIFIX_BY_ACTION_NAME['Post memo']:
            raise TypeError('This is not a post memo!')

        line_sender = get_name_from_address(self.sender) + '@...' + self.sender[-6:] \
                      + ' posted at ' + datetime.fromtimestamp(int(self.transaction_time)).strftime('%Y-%m-%d %H:%M:%S')
        line_message = self.message

        return line_sender + '\n' + line_message + '\n'

    def content_like(self):
        if self.prefix!=PRIFIX_BY_ACTION_NAME['Like / tip memo']:
            raise TypeError('This is not a like / tip memo!')

        # if self.like_memo is None:
        #     presentation_of_liked_memo='Not a memo'
        # elif self.like_memo.sender=='':
        #     presentation_of_liked_memo='No sender address'
        # else:
        #     presentation_of_liked_memo=get_name_from_address(self.like_memo.sender)+'\'s post'
        #
        # line='Liked '\
        #      +presentation_of_liked_memo \
        #      +' ('+self.txhash+')'\
        #      +'- '+ '{0:,d}'.format(self._tip_aount) + ' satoshis' \
        #      +' at '+datetime.fromtimestamp(int(self.time)).strftime('%Y-%m-%d %H:%M:%S')

        line='Liked '\
             +' ('+self.txhash_of_target_memo + ')'\
             +' at '+datetime.fromtimestamp(int(self.transaction_time)).strftime('%Y-%m-%d %H:%M:%S')

        return line

    def content_reply(self):
        if self.prefix!=PRIFIX_BY_ACTION_NAME['Reply to memo']:
            raise TypeError('This is not a Reply to memo!')

        line_sender = get_name_from_address(self.sender) + '@...' + self.sender[-6:] \
                      + ' replied to '\
                      + 'memo({})'.format(self.txhash_of_target_memo)\
                      + ' at ' \
                      + datetime.fromtimestamp(int(self.transaction_time)).strftime('%Y-%m-%d %H:%M:%S')
        line_message = self.message

        return line_sender + '\n' + line_message + '\n'

    def __create_values(self):
        """
        Create values hex string based on action type and action values.
        """
        if self._prefix==PRIFIX_BY_ACTION_NAME['Set name']:
            name_bytes=self.name.encode()
            len_name_bytes = len(name_bytes).to_bytes(1, byteorder='little')
            self._values=(len_name_bytes + name_bytes).hex()
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Post memo']:
            message_bytes=self.message.encode()
            len_message_bytes = len(message_bytes).to_bytes(1, byteorder='little')
            self._values=(len_message_bytes + message_bytes).hex()
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Like / tip memo']:
            tx_hash_bytes=bytes.fromhex(self.txhash_of_target_memo)
            len_tx_hash_bytes=len(tx_hash_bytes).to_bytes(1,byteorder='little')
            self._values=(len_tx_hash_bytes+bytes(reversed(tx_hash_bytes))).hex()
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Follow user']:
            address_bytes=address_to_public_key_hash(self.address)
            len_address_bytes=len(address_bytes).to_bytes(1,byteorder='little')
            self._values=(len_address_bytes+address_bytes).hex()
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Unfollow user']:
            address_bytes=address_to_public_key_hash(self.address)
            len_address_bytes=len(address_bytes).to_bytes(1,byteorder='little')
            self._values=(len_address_bytes+address_bytes).hex()
        elif self._prefix==PRIFIX_BY_ACTION_NAME['Reply to memo']:
            tx_hash_bytes=bytes.fromhex(self.txhash_of_target_memo)
            len_tx_hash_bytes=len(tx_hash_bytes).to_bytes(1,byteorder='little')
            message_bytes=self.message.encode()
            len_message_bytes = len(message_bytes).to_bytes(1, byteorder='little')
            self._values=(len_tx_hash_bytes + bytes(reversed(tx_hash_bytes))
                         +len_message_bytes + message_bytes).hex()


    @classmethod
    def set_name(cls,name,sender):
        memo=Memo()
        memo.sender=sender
        memo.prefix=PRIFIX_BY_ACTION_NAME['Set name']
        memo.name=name
        memo.__create_values()
        return memo

    @classmethod
    def post_memo(cls,message,sender):
        memo=Memo()
        memo.sender=sender
        memo.prefix=PRIFIX_BY_ACTION_NAME['Post memo']
        memo.message=message
        memo.__create_values()
        return memo

    @classmethod
    def like_memo(cls, liked_memo, sender, tip_amount=None, sender_of_liked_memo=None):
        """
        :type liked_memo: ``Memo``,  ``str``

        """

        memo=Memo()
        memo.sender=sender
        memo.prefix=PRIFIX_BY_ACTION_NAME['Like / tip memo']
        if isinstance(liked_memo,Memo):
            memo.liked_memo=liked_memo
            memo.txhash_of_target_memo=memo.liked_memo.transaction_hash
            memo._sender_of_liked_memo=memo.liked_memo.sender
        elif isinstance(liked_memo,str):
            memo.txhash_of_target_memo=liked_memo
            memo._sender_of_liked_memo=sender_of_liked_memo
        else:
            raise TypeError('Wrong liked_memo({}) type!'.format(liked_memo))
        memo.__create_values()

        if not isinstance(tip_amount,int) and tip_amount is not None:
            raise TypeError('tip_amount should be a string!')
        elif isinstance(tip_amount,int) and tip_amount<546:
            raise ValueError('Min transaction amount is 546')
        elif isinstance(tip_amount,int):
            memo._tip_amount=int(tip_amount)
            memo.__create_transfer()
        return memo

    def __create_transfer(self):
        if self._tip_amount!=0:
            self.transfer=[(self._sender_of_liked_memo, self._tip_amount, 'satoshi')]

    @classmethod
    def follow(cls,user_address,sender):
        memo=Memo()
        memo.sender=sender
        memo.prefix=PRIFIX_BY_ACTION_NAME['Follow user']
        memo.address=address_to_public_key_hash(user_address).hex()
        memo.__create_values()
        return memo

    @classmethod
    def unfollow(cls,user_address,sender):
        memo=Memo()
        memo.sender=sender
        memo.prefix=PRIFIX_BY_ACTION_NAME['Unfollow user']
        memo.address=address_to_public_key_hash(user_address).hex()
        memo.__create_values()
        return memo

    @classmethod
    def reply_to_memo(cls,target_memo,message,sender):
        memo=Memo()
        memo.sender=sender
        memo.prefix=PRIFIX_BY_ACTION_NAME['Reply to memo']
        if isinstance(target_memo,Memo):
            memo.txhash_of_target_memo=target_memo.transaction_hash
        elif isinstance(target_memo,str):
            memo.txhash_of_target_memo=target_memo
        else:
            raise TypeError('Wrong type of target_memo')
        memo.message=message
        memo.__create_values()
        return memo

    def create_signed_transaction(self,private_key,leftover=None):
        """

        :param private_key:
        :param leftover: address for receiving left bch. Default is sender.
        :return:
        """
        pk=PrivateKey(private_key)
        if pk.address != self.sender:
            raise ValueError('Wrong Private Key!')

        prefix_bytes=bytes.fromhex(self.prefix)
        len_prefix_bytes=len(prefix_bytes).to_bytes(1,byteorder='little')
        values_bytes=bytes.fromhex(self._values)

        data_bytes=len_prefix_bytes+prefix_bytes+values_bytes

        pk.get_unspents()
        self.signed_transaction=pk.create_transaction(message=data_bytes,outputs=self.transfer,fee=MIN_TRANSFER_FEE,leftover=leftover)

        return self.signed_transaction

    def send_transaction(self):
        NetworkAPI.broadcast_tx(self.signed_transaction)
        return bitcash.wallet.calc_txid(self.signed_transaction)

    @classmethod
    def from_txhash(cls,txhash):
        tx_dict=NetworkAPI.get_tx(txhash)
        return Memo.form_transaction_dict(tx_dict)

class NotMemoTransaction(Exception):
    pass

def get_name_from_address(address):
    if address in USER_NAME_DICT:
        return USER_NAME_DICT[address]
    else:
        return address[-6:]


from distutils.core import setup

setup(
    name='bchmemo',
    version='0.0.3',
    packages=['bchmemo','bchmemo.bitcash_modified'],
    url='https://github.com/remem9527/memo',
    license='MIT',
    author='remem9527',
    author_email='wszd163@163.com',
    description='A python package for memo. Memo is an on-chain social network built on Bitcoin Cash',
    long_description=open('README.rst', 'r').read(),
    keywords=[
        'bitcoincash',
        'social network',
        'memo',
        'on-chain data',
        'OP_RETURN',
    ],

    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],

    install_requires=['bitcash']

)
